// By Jeroen Taverne

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>

#define MAPPER_DETECT 0

#ifdef MAPPER_DETECT
#include "SHA1.c"
#include "guessrom.c"
#else
#define MAP_NOT_DETECTED 255
#endif

#define FILENAME_SIZE 256
#define ENTRY_SIZE 96
#define NAME_SIZE 80
#define SIZE_POS 80
#define MAPPER_POS 84
#define GEN_POS 86

const char *mapper_table[] =
{
	"generic8",
	"generic16",
	"konami5",
	"konami4",
	"ascii8",
	"ascii16",
	"gamemaster2",
	"fmpac",
	"ascii16ex",
	"rtype",
	"nextor_ram",
	"disk",
	"plain0000",
	"plain4000",
	"plain8000",
	"neo16",
	"neo8",
	"konamiultimate",
	0
};

const char *msxgen_table[] =
{
	"msx1",
	"msx2",
	"msx2+",
	"turbor",
	0
};

static uint8_t search_table(char *text, const char *table[])
{
	uint8_t i = 0;

	while (table[i])
	{
		char *p = strstr(text, "_");
		if (p)
		{
			int len = strlen(table[i]);
			p = strstr(p, table[i]);
			if (p && (p[len] == '_' || p[len] == '.'))
			{
				return i;
			}
		}
		i++;
	}
	return 255;
}

int main(int argc, char *argv[])
{
	DIR* dirp;
	struct dirent* direntp;
	uint32_t filecount = 0;
	uint32_t skipcount = 0;

	FILE *f_input, *f_output;

	if (argc != 3)
	{
		fprintf(stderr, "Usage: %s <ROMs path> <output file>\n", argv[0]);
		return -1;
	}

	chdir(argv[1]);

	dirp = opendir(".");
	if (dirp == NULL)
	{
		printf("Can't open %s\n", argv[1]);
		exit(-1);
	}

	f_output = fopen(argv[2], "wb");
	if (f_output == NULL)
	{
		fprintf(stderr, "%s: can't open %s for writing\n", argv[0], argv[2]);
		return -1;
	}
	fwrite("MSXPICO_ROM_CAT", 16, 1, f_output);

	while (1)
	{
		direntp = readdir(dirp);
		if (direntp == NULL) break;

		char name[FILENAME_SIZE];
		char name_lc[FILENAME_SIZE];
		strncpy(name, direntp->d_name, FILENAME_SIZE);
		for (int i=0; i<FILENAME_SIZE; i++)
		{
			name_lc[i] = tolower(name[i]);
		}

		char *p = strstr(name_lc, ".rom");
		if (p && strlen(p) == 4)
		{
			int filesize = 0;

			uint8_t mapper = search_table(name_lc, mapper_table);
			uint8_t msxgen = search_table(name_lc, msxgen_table);

			char *filename = direntp->d_name;
			f_input = fopen(direntp->d_name, "rb");
			if (f_input == NULL)
			{
				fprintf(stderr, "%s: can't open %s for reading\n", argv[0], filename);
				return -1;
			}
			fseek(f_input, 0L, SEEK_END);
			filesize = ftell(f_input);
			fseek(f_input, 0L, SEEK_SET);
			filesize = (filesize + 15) & 0xfffffff0;
			uint8_t *buffer = calloc(filesize, 1);
			fread(buffer, filesize, 1, f_input);
			fclose(f_input);

#if MAPPER_DETECT
			if (mapper >= MAP_NOT_DETECTED)
			{
				mapper = guess_rom(buffer, filesize);
			}
#endif

			char *p = strstr(name, "_");
			if (p)
			{
				*p = 0;
			}

			if (mapper < MAP_NOT_DETECTED && msxgen <= 3)
			{
				printf("Name: %s | Mapper:%s | Generation: %s\n", name, mapper_table[mapper], msxgen_table[msxgen]);

				uint8_t entry[ENTRY_SIZE];
				memset(entry, 0, ENTRY_SIZE);
				strncpy((char*)entry, name, 0x40);
				entry[SIZE_POS+0] = (filesize >> 0) & 0xff;
				entry[SIZE_POS+1] = (filesize >> 8) & 0xff;
				entry[SIZE_POS+2] = (filesize >> 16) & 0xff;
				entry[SIZE_POS+3] = (filesize >> 24) & 0xff;
				entry[MAPPER_POS] = mapper;
				entry[GEN_POS] = msxgen;
				fwrite(entry, ENTRY_SIZE, 1, f_output);
				fwrite(buffer, filesize, 1, f_output);
				
				filecount++;
			}
			else
			{
				printf("Name: %s Mapper or generation not specified, skipped!\n", name);
				
				skipcount++;
			}

			free(buffer);
		}
	}

	uint8_t entry[ENTRY_SIZE];
	memset(entry, 0, ENTRY_SIZE);
	fwrite(entry, ENTRY_SIZE, 1, f_output);
	
	printf("Files added : %u\n", filecount);
	printf("Files skipped : %u\n", skipcount);
	printf("Output file size : %ld\n", ftell(f_output));

	fclose(f_output);
	closedir( dirp );

	return 0;
}
